#ifdef WIN32
#include <io.h>
#include <stdlib.h>
#else
#include <sys/types.h>
#include <dirent.h>
#endif
#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <string>
#include <vector>
#include "gigabase.h"

USE_GIGABASE_NAMESPACE

#include "globals.h"
#include "utm_cl.h"
#include "bzlib.h"

REGISTER_UNASSIGNED(iptraffic_raw);

iptraffic_raw_utm::iptraffic_raw_utm ()
{ something1 = 0;
  sourceaddr = 0;
  destaddr = 0;
  nexthop = 0;
  something2 = 0;
  packets = 0;
  octets = 0;
  first = 0;
  last = 0;
  src_port = 0;
  dest_port = 0;
  tcp_flags = 0;
  proto = 0;
  tos = 0;
  something3 = 0;
  src_as = 0;
  dst_as = 0;
  something4 = 0;
  slink_id = 0;
  account_id = 0;
  sourceaddr1 = 0;
  t_class = 0;
  timestamp = 0;
  nfgen_addr = 0;
}

iptraffic_raw_utm::~iptraffic_raw_utm ()
{}

void iptraffic_raw_utm::Load (FILE *data)
{ unsigned char buffer[4];

  something1 = 0;
  sourceaddr = 0;
  destaddr = 0;
  nexthop = 0;
  something2 = 0;
  packets = 0;
  octets = 0;
  first = 0;
  last = 0;
  src_port = 0;
  dest_port = 0;
  tcp_flags = 0;
  proto = 0;
  tos = 0;
  something3 = 0;
  src_as = 0;
  dst_as = 0;
  something4 = 0;
  slink_id = 0;
  account_id = 0;
  sourceaddr1 = 0;
  t_class = 0;
  timestamp = 0;
  nfgen_addr = 0;
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  something1 = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  sourceaddr = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  destaddr = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  nexthop = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  something2 = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  packets = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  octets = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  first = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  last = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,2,data) != 2) throw ioerror (errno);
  src_port = (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,2,data) != 2) throw ioerror (errno);
  dest_port = (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,2,data) != 2) throw ioerror (errno);
  tcp_flags = (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,1,data) != 1) throw ioerror (errno);
  proto = buffer[0];
  if (fread (&buffer,1,1,data) != 1) throw ioerror (errno);
  tos = buffer[0];
  if (fread (&buffer,1,2,data) != 2) throw ioerror (errno);
  something3 = (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,2,data) != 2) throw ioerror (errno);
  src_as = (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,2,data) != 2) throw ioerror (errno);
  dst_as = (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,2,data) != 2) throw ioerror (errno);
  something4 = (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  slink_id = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  account_id = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  sourceaddr1 = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  t_class = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  timestamp = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
  if (fread (&buffer,1,4,data) != 4) throw ioerror (errno);
  nfgen_addr = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
}

UTMRawBaseQuery::UTMRawBaseQuery ()
{ path = "";
  rows_count = 0;
  range_low = -1;
  range_high = -1;
}

UTMRawBaseQuery::UTMRawBaseQuery (std::string& datapath)
{ path = datapath;
  rows_count = 0;
  range_low = -1;
  range_high = -1;
}

UTMRawBaseQuery::UTMRawBaseQuery (char const *datapath)
{ path = (datapath!=NULL)?datapath:"";
  rows_count = 0;
  range_low = -1;
  range_high = -1;
}

UTMRawBaseQuery::~UTMRawBaseQuery ()
{}

void GetTempDirectory (std::string& d)
{ d =
#ifdef WIN32
  getenv ("TEMP")
#else
  "/var/tmp"
#endif
  ;
  d += PathDivider;
}

void UTMRawBaseQuery::Execute ()
{
#ifdef WIN32
  struct _finddata_t fd;
  int fh;
#else
  DIR *d;
  struct dirent *e;
#endif
  int ctr;
  char goodfile;
  int min_ts_pos,min_ts,ts;
  std::string fname;
  std::string tempdir;
  std::vector<int> found_timestamps;
  std::vector<int> timestamps_to_process;
  char buffer[32];
  FILE *datafile,*tmpfile;
  iptraffic_raw_utm rec;
  char record_ok;
  BZFILE *bzf;
  char packed,unpacked;
  int bz_err,bz_bytes_num,bz_bytes_unused;
  char *bz_buffer,*bz_unused;
  void *bz_unusedtmpv;
  unsigned char *bz_unusedtmp;

  rows_count = 0;
#ifdef WIN32
  fname = path + "iptraffic_raw_*.*";
  fh = _findfirst (fname.c_str (),&fd);
  if (fh != -1)
#else
  d = opendir (path.c_str ());
  if (d != NULL)
#endif
  { // Finding timestamps
#ifdef WIN32
    do
    { if (strlen ((char *)(&fd.name)) > 14)
      { ctr = 14;
        goodfile = 1;
        while (fd.name[ctr]!=0)
        { if (!isdigit(fd.name[ctr]))
          { if (fd.name[ctr] != '.') goodfile = 0;
            break;
          }
          ctr++;
        }
        goodfile &= ((strcmp ((char *)(&(fd.name[ctr])),".utm.bz2") == 0) || (strcmp ((char *)(&(fd.name[ctr])),".utm") == 0));
        if (goodfile)
        { ts = 0;
          ctr = 14;
          while ((fd.name[ctr]!=0) && (fd.name[ctr]!='.'))
          { if (isdigit (fd.name[ctr]))
            { if (isdigit (fd.name[ctr-1])) ts *= 10;
              ts += fd.name[ctr]-'0';
            }
            else break;
            ctr++;
          }
          found_timestamps.insert (found_timestamps.begin (),ts);
        }
      }
    }
    while (_findnext (fh,&fd) == 0);
    _findclose (fh);
#else
    while ((e = readdir (d)) != NULL)
    { if (((e->d_type == DT_REG) || (e->d_type == DT_LNK)) && (e->d_namlen > 14))
      { if (strncmp ((char *)(&(e->d_name[0])),"iptraffic_raw_",14) == 0)
        { ctr = 14;
          goodfile = 1;
          while (ctr < e->d_namlen)
          { if (!isdigit(e->d_name[ctr]))
            { if (e->d_name[ctr] != '.') goodfile = 0;
              break;
            }
            ctr++;
          }
          goodfile &= ((strcmp ((char *)(&(e->d_name[ctr])),".utm.bz2") == 0) || (strcmp ((char *)(&(e->d_name[ctr])),".utm") == 0));
          if (goodfile)
          { ts = 0;
            ctr = 14;
            while ((ctr < e->d_namlen) && (e->d_name[ctr]!='.'))
            { if (isdigit (e->d_name[ctr]))
              { if (isdigit (e->d_name[ctr-1])) ts *= 10;
                ts += e->d_name[ctr]-'0';
              }
              else break;
              ctr++;
            }
            found_timestamps.insert (found_timestamps.begin (),ts);
          }
        }
      }
    }
    closedir (d);
#endif
    if (found_timestamps.size () > 0)
    { if (found_timestamps.size () == 1)
       timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[0]);
      else
      { // Sorting timestamps, to further correct processing
        for (int j = 0;j < found_timestamps.size ();j++)
        { min_ts = found_timestamps[j];
          min_ts_pos = j;
          for (int k = j;k < found_timestamps.size ();k++)
          { if (found_timestamps[k] < min_ts)
            { min_ts = found_timestamps[k];
              min_ts_pos = k;
            }
          }
          int tmp = found_timestamps[j];
          found_timestamps[j] = found_timestamps[min_ts_pos];
          found_timestamps[min_ts_pos] = tmp;
        }
        if ((range_low != -1) && (range_high != -1))
        { // Finding files with data in specified range
          for (int j = 0;j < found_timestamps.size ();j++)
          { if ((found_timestamps[j] >= range_low) &&
                (found_timestamps[j] <= range_high))
             timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[j]);
          }
          for (int j = found_timestamps.size () - 1,in_range = 0;j >= 0;j--)
          { if ((found_timestamps[j] <= range_high) && (in_range == 0)) in_range = 1;
            if ((found_timestamps[j] <= range_low) && (in_range == 1))
            { timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[j]);
              break;
            }
          }
        }
        else
        { if ((range_low == -1) && (range_high == -1))
          { for (int j = 0;j < found_timestamps.size ();j++)
             timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[j]);
          }
          else
          { if (range_low == -1)
            { for (int j = 0;j < found_timestamps.size ();j++)
              { if (found_timestamps[j] <= range_high)
                 timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[j]);
              }
            }
            else
            { if (range_high == -1)
              { for (int j = 0;j < found_timestamps.size ();j++)
                { if (found_timestamps[j] >= range_low)
                   timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[j]);
                }
                for (int j = found_timestamps.size () - 1,in_range = 0;j >= 0;j--)
                { if (found_timestamps[j] <= range_low)
                  { timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[j]);
                    break;
                  }
                }
              }
            }
          }
        }
      }
      if (timestamps_to_process.size () > 0)
      { // Processing data files
        this->ShowProgress (0,NULL,0,0);
        for (int j = 0,k = timestamps_to_process.size ();j < k;j++)
        { if (stopped) break;
          sprintf ((char *)(&buffer),"%d",timestamps_to_process[j]);
          fname = path + "iptraffic_raw_" + (char *)(&buffer) + ".utm.bz2";
          datafile = fopen (fname.c_str (),"rb");
          if (datafile != NULL)
          { // Unpacking
            fname = MakeTempName ();
//            fname += "iptraffic_raw_";
//            fname += (char *)(&buffer);
//            fname += ".utm";
            packed = 1;
            unpacked = 0;
            bz_buffer = new char[8192];
            bz_unused = new char[BZ_MAX_UNUSED];
            bz_bytes_unused = 0;
            tmpfile = fopen (fname.c_str (),"wb");
            if (tmpfile != NULL)
            { unpacked = 0;
              while (1)
              { bzf = BZ2_bzReadOpen (&bz_err,datafile,0,0,bz_unused,bz_bytes_unused);
                if (bzf == NULL || bz_err != BZ_OK) break;
                while (bz_err == BZ_OK)
                { bz_bytes_num = BZ2_bzRead (&bz_err,bzf,bz_buffer,8192);
                  if ((bz_err == BZ_OK || bz_err == BZ_STREAM_END) && bz_bytes_num > 0)
                   fwrite (bz_buffer,sizeof (unsigned char),bz_bytes_num,tmpfile);
                  if (ferror(datafile)) break;
                }
                if (bz_err != BZ_STREAM_END) break;
                BZ2_bzReadGetUnused (&bz_err,bzf,&bz_unusedtmpv,&bz_bytes_unused);
                bz_unusedtmp = (unsigned char *)bz_unusedtmpv;
                for (int j = 0;j < bz_bytes_unused;j++) bz_unused[j] = bz_unusedtmp[j];
                BZ2_bzReadClose (&bz_err,bzf);
                if (bz_bytes_unused == 0 && feof(datafile))
                { unpacked = 1;
                  break;
                }
              }
              fclose (datafile);
              fclose (tmpfile);
              if (!unpacked) unlink (fname.c_str ());
              else datafile = fopen (fname.c_str (),"rb");
            }
            delete [] bz_unused;
            delete [] bz_buffer;
          }
          else
          { fname = path + "iptraffic_raw_" + (char *)(&buffer) + ".utm";
            datafile = fopen (fname.c_str (),"rb");
            packed = 0;
            unpacked = 1;
          }
          if (datafile != NULL)
          { while (!feof (datafile))
            { record_ok = 1;
              try
              { rec.Load (datafile);
              }
              catch (iptraffic_raw_utm::ioerror& e)
              { record_ok = 0;
              }
              if (record_ok)
              { if ((this->Filter (rec)) && (this->ProcessRecord (rec) == 0)) break;
              }
              else break;
            }
            fclose (datafile);
            this->ShowProgress (1,fname.c_str (),j+1,k);
            if (packed) unlink (fname.c_str ());
          }
        }
        this->ShowProgress (2,NULL,0,0);
      }
    }
  }
}

void UTMRawBaseQuery::ShowProgress (int state,char const *currfile,int currfile_num,int total_files)
{}

int UTMRawBaseQuery::Filter (iptraffic_raw_utm& trec)
{ if ((this->range_low != -1) && (this->range_high != -1))
   return ((trec.timestamp >= this->range_low) && (trec.timestamp < this->range_high));
  if (this->range_low != -1)
   return (trec.timestamp >= this->range_low);
  if (this->range_high != -1)
   return (trec.timestamp < this->range_high);
}

int UTMRawBaseQuery::ProcessRecord (iptraffic_raw_utm& trec)
{ return 1;
}

void UTMRawBaseQuery::SetTimeRange (int r_low,int r_high)
{ if ((r_low != -1) && (r_high != -1))
  { if (r_low < r_high)
    { range_low = r_low;
      range_high = r_high;
    }
    else
    { range_low = r_high;
      range_high = r_low;
    }
  }
  else
  { if (r_low != -1)
    { range_low = r_low;
      range_high = -1;
    }
    else
    { if (r_high != -1)
      { range_low = -1;
        range_high = r_high;
      }
    }
  }
}

UTMGBBaseQuery::UTMGBBaseQuery ()
{ path = "";
  rows_count = 0;
  range_low = -1;
  range_high = -1;
}

UTMGBBaseQuery::UTMGBBaseQuery (std::string& gbpath)
{ path = gbpath;
  rows_count = 0;
  range_low = -1;
  range_high = -1;
}

UTMGBBaseQuery::UTMGBBaseQuery (char const *gbpath)
{ path = (gbpath!=NULL)?gbpath:"";
  rows_count = 0;
  range_low = -1;
  range_high = -1;
}

UTMGBBaseQuery::~UTMGBBaseQuery ()
{}

void UTMGBBaseQuery::Execute ()
{
#ifdef WIN32
  struct _finddata_t fd;
  int fh;
#else
  DIR *d;
  struct dirent *e;
#endif
  char goodfile;
  int ctr;
  int min_ts_pos,min_ts,ts;
  std::string fname;
  std::string tempdir;
  std::vector<int> found_timestamps;
  std::vector<int> timestamps_to_process;
  char buffer[32];
  dbDatabase *tdb;
  dbCursor<iptraffic_raw> *tdc;
  dbQuery *tq;
  FILE *datafile,*tmpfile;
  BZFILE *bzf;
  char packed,unpacked;
  int bz_err,bz_bytes_num,bz_bytes_unused;
  char *bz_buffer,*bz_unused;
  void *bz_unusedtmpv;
  unsigned char *bz_unusedtmp;

  rows_count = 0;
#ifdef WIN32
  fname = path + "iptraffic_raw_*.*";
  fh = _findfirst (fname.c_str (),&fd);
  if (fh != -1)
#else
  d = opendir (path.c_str ());
  if (d != NULL)
#endif
  { // Finding timestamps
#ifdef WIN32
    do
    { if (strlen ((char *)(&fd.name)) > 14)
      { ctr = 14;
        goodfile = 1;
        while (fd.name[ctr]!=0)
        { if (!isdigit(fd.name[ctr]))
          { if (fd.name[ctr] != '.') goodfile = 0;
            break;
          }
          ctr++;
        }
        goodfile &= ((strcmp ((char *)(&(fd.name[ctr])),".dbs.bz2") == 0) || (strcmp ((char *)(&(fd.name[ctr])),".dbs") == 0));
        if (goodfile)
        { ts = 0;
          ctr = 14;
          while ((fd.name[ctr]!=0) && (fd.name[ctr]!='.'))
          { if (isdigit (fd.name[ctr]))
            { if (isdigit (fd.name[ctr-1])) ts *= 10;
              ts += fd.name[ctr]-'0';
            }
            else break;
            ctr++;
          }
          found_timestamps.insert (found_timestamps.begin (),ts);
        }
      }
    }
    while (_findnext (fh,&fd) == 0);
    _findclose (fh);
#else
    while ((e = readdir (d)) != NULL)
    { if (((e->d_type == DT_REG) || (e->d_type == DT_LNK)) && (e->d_namlen > 14))
      { if (strncmp (&(e->d_name[0]),"iptraffic_raw_",14) == 0)
        { ctr = 14;
          goodfile = 1;
          while (ctr < e->d_namlen)
          { if (!isdigit(e->d_name[ctr]))
            { if (e->d_name[ctr] != '.') goodfile = 0;
              break;
            }
            ctr++;
          }
          goodfile &= ((strcmp ((char *)(&(e->d_name[ctr])),".dbs.bz2") == 0) || (strcmp ((char *)(&(e->d_name[ctr])),".dbs") == 0));
          if (goodfile)
          { ts = 0;
            ctr = 14;
            while ((ctr < e->d_namlen) && (e->d_name[ctr]!='.'))
            { if (isdigit (e->d_name[ctr]))
              { if (isdigit (e->d_name[ctr-1])) ts *= 10;
                ts += e->d_name[ctr]-'0';
              }
              else break;
              ctr++;
            }
            found_timestamps.insert (found_timestamps.begin (),ts);
          }
        }
      }
    }
    closedir (d);
#endif
    if (found_timestamps.size () > 0)
    { tdb = new dbDatabase(dbDatabase::dbReadOnly,4096,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
      tdc = new dbCursor<iptraffic_raw> (tdb);
      tq = new dbQuery ("");
      if (found_timestamps.size () == 1)
       timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[0]);
      else
      { // Sorting timestamps, to further correct processing
        for (int j = 0;j < found_timestamps.size ();j++)
        { min_ts = found_timestamps[j];
          min_ts_pos = j;
          for (int k = j;k < found_timestamps.size ();k++)
          { if (found_timestamps[k] < min_ts)
            { min_ts = found_timestamps[k];
              min_ts_pos = k;
            }
          }
          int tmp = found_timestamps[j];
          found_timestamps[j] = found_timestamps[min_ts_pos];
          found_timestamps[min_ts_pos] = tmp;
        }
        if ((range_low != -1) && (range_high != -1))
        { // Finding files with data in specified range
          for (int j = 0;j < found_timestamps.size ();j++)
          { if ((found_timestamps[j] >= range_low) &&
                (found_timestamps[j] <= range_high))
             timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[j]);
          }
          for (int j = 0,in_range = 0;j < found_timestamps.size ();j++)
          { if ((found_timestamps[j] >= range_low) && (in_range == 0)) in_range = 1;
            if ((found_timestamps[j] >= range_high) && (in_range == 1))
            { timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[j]);
              break;
            }
          }
        }
        else
        { if ((range_low == -1) && (range_high == -1))
          { for (int j = 0;j < found_timestamps.size ();j++)
             timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[j]);
          }
          else
          { if (range_low == -1)
            { for (int j = 0;j < found_timestamps.size ();j++)
              { if (found_timestamps[j] <= range_high)
                 timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[j]);
              }
              for (int j = 0;j < found_timestamps.size ();j++)
              { if (found_timestamps[j] >= range_high)
                { timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[j]);
                  break;
                }
              }
            }
            else
            { if (range_high == -1)
              { for (int j = 0;j < found_timestamps.size ();j++)
                { if (found_timestamps[j] >= range_low)
                   timestamps_to_process.insert (timestamps_to_process.begin (),found_timestamps[j]);
                }
              }
            }
          }
        }
      }
      if (timestamps_to_process.size () > 0)
      { // GetTempDirectory (tempdir);
        this->ShowProgress (0,NULL,0,0);
        // Processing data files
        for (int j = 0,k = timestamps_to_process.size ();j < k;j++)
        { if (stopped) break;
          sprintf ((char *)(&buffer),"%d",timestamps_to_process[j]);
          fname = path + "iptraffic_raw_" + (char *)(&buffer) + ".dbs.bz2";
          datafile = fopen (fname.c_str (),"rb");
          if (datafile != NULL)
          { // Unpacking
            //fname = tempdir;
            //fname += "iptraffic_raw_";
            //fname += (char *)(&buffer);
            //fname += ".dbs";
            fname = MakeTempName ();
            packed = 1;
            unpacked = 0;
            bz_buffer = new char[8192];
            bz_unused = new char[BZ_MAX_UNUSED];
            bz_bytes_unused = 0;
            tmpfile = fopen (fname.c_str (),"wb");
            if (tmpfile != NULL)
            { unpacked = 0;
              while (1)
              { bzf = BZ2_bzReadOpen (&bz_err,datafile,0,0,bz_unused,bz_bytes_unused);
                if (bzf == NULL || bz_err != BZ_OK) break;
                while (bz_err == BZ_OK)
                { bz_bytes_num = BZ2_bzRead (&bz_err,bzf,bz_buffer,8192);
                  if ((bz_err == BZ_OK || bz_err == BZ_STREAM_END) && bz_bytes_num > 0)
                   fwrite (bz_buffer,sizeof (unsigned char),bz_bytes_num,tmpfile);
                  if (ferror(datafile)) break;
                }
                if (bz_err != BZ_STREAM_END) break;
                BZ2_bzReadGetUnused (&bz_err,bzf,&bz_unusedtmpv,&bz_bytes_unused);
                bz_unusedtmp = (unsigned char *)bz_unusedtmpv;
                for (int j = 0;j < bz_bytes_unused;j++) bz_unused[j] = bz_unusedtmp[j];
                BZ2_bzReadClose (&bz_err,bzf);
                if (bz_bytes_unused == 0 && feof(datafile))
                { unpacked = 1;
                  break;
                }
              }
              fclose (datafile);
              fclose (tmpfile);
              if (!unpacked) unlink (fname.c_str ());
            }
            delete [] bz_unused;
            delete [] bz_buffer;
          }
          else
          { fname = path + "iptraffic_raw_" + (char *)(&buffer) + ".dbs";
            packed = 0;
            unpacked = 1;
          }
          // if (unpacked)
#ifdef dbException
          try
          {
#endif
            tdb->open (fname.c_str ());
            if (tdb->isOpen ())
            { if ((this->rows_count = tdc->select (*tq)) > 0)
              { do
                { if ((this->Filter (*tdc)) && (this->ProcessRow (*tdc) == 0)) break;
                }
                while (tdc->next ());
              }
              tdb->close ();
              this->ShowProgress (1,fname.c_str (),j+1,k);
            }
#ifdef dbException
          }
          catch (dbException& e)
          {
          }
#endif
          if (packed) unlink (fname.c_str ());
        }
        this->ShowProgress (2,NULL,0,0);
      }
      delete tq;
      delete tdc;
      delete tdb;
    }
  }
}

void UTMGBBaseQuery::ShowProgress (int state,char const *currdb_name,int currdb_num,int total_dbs)
{}

int UTMGBBaseQuery::Filter (dbCursor<iptraffic_raw> &tdc)
{ if ((this->range_low != -1) && (this->range_high != -1))
   return ((tdc->timestamp >= this->range_low) && (tdc->timestamp < this->range_high));
  if (this->range_low != -1)
   return (tdc->timestamp >= this->range_low);
  if (this->range_high != -1)
   return (tdc->timestamp < this->range_high);
}

int UTMGBBaseQuery::ProcessRow (dbCursor<iptraffic_raw> &tdc)
{ return 1;
}

void UTMGBBaseQuery::SetTimeRange (int r_low,int r_high)
{ if ((r_low != -1) && (r_high != -1))
  { if (r_low < r_high)
    { range_low = r_low;
      range_high = r_high;
    }
    else
    { range_low = r_high;
      range_high = r_low;
    }
  }
  else
  { if (r_low != -1)
    { range_low = r_low;
      range_high = -1;
    }
    else
    { if (r_high != -1)
      { range_low = -1;
        range_high = r_high;
      }
    }
  }
}

UTMRawToGB::UTMRawToGB () : UTMRawBaseQuery ()
{ commitcounter = 0;
  db_to_name = "";
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

UTMRawToGB::UTMRawToGB (std::string& datapath,std::string& _db_to_name) : UTMRawBaseQuery (datapath)
{ commitcounter = 0;
  db_to_name = _db_to_name;
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

UTMRawToGB::UTMRawToGB (std::string& datapath,char const *_db_to_name) : UTMRawBaseQuery (datapath)
{ commitcounter = 0;
  db_to_name = _db_to_name;
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

UTMRawToGB::UTMRawToGB (char const *datapath,std::string& _db_to_name) : UTMRawBaseQuery (datapath)
{ commitcounter = 0;
  db_to_name = _db_to_name;
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

UTMRawToGB::UTMRawToGB (char const *datapath,char const *_db_to_name) : UTMRawBaseQuery (datapath)
{ commitcounter = 0;
  db_to_name = _db_to_name;
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

UTMRawToGB::~UTMRawToGB ()
{ if (this->db_to->isOpen ()) this->db_to->close ();
  delete this->db_to;
}

void UTMRawToGB::Execute ()
{ this->db_to->open (this->db_to_name.c_str ());
  UTMRawBaseQuery::Execute ();
  this->db_to->commit ();
  this->db_to->close ();
}

int UTMRawToGB::ProcessRecord (iptraffic_raw_utm& trec)
{ iptraffic_raw rec;

  if (stopped) return 0;
  rec.slink_id = trec.slink_id;
  rec.srcaddr = trec.sourceaddr;
  rec.dstaddr = trec.destaddr;
  rec.nexthop = trec.nexthop;
  rec.input = 0;
  rec.output = 0;
  rec.dPkts = trec.packets;
  rec.dOctets = trec.octets;
  rec.First = trec.first;
  rec.Last = trec.last;
  rec.srcport = trec.src_port;
  rec.dstport = trec.dest_port;
  rec.pad = 0;
  rec.tcp_flags = trec.tcp_flags;
  rec.prot = trec.proto;
  rec.tos = trec.tos;
  rec.src_as = trec.src_as;
  rec.dst_as = trec.dst_as;
  rec.src_mask = 0;
  rec.dst_mask = 0;
  rec.t_class = trec.t_class;
  rec.account_id = trec.account_id;
  rec.timestamp = trec.timestamp;
  this->db_to->insert (rec);
  this->commitcounter++;
  if (this->commitcounter == 10000)
  { this->db_to->commit ();
    this->commitcounter = 0;
  }
  return 1;
}

UTMGBToGB::UTMGBToGB () : UTMGBBaseQuery ()
{ commitcounter = 0;
  db_to_name = "";
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

UTMGBToGB::UTMGBToGB (std::string& gbpath,std::string& _db_to_name) : UTMGBBaseQuery (gbpath)
{ commitcounter = 0;
  db_to_name = _db_to_name;
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

UTMGBToGB::UTMGBToGB (std::string& gbpath,char const *_db_to_name) : UTMGBBaseQuery (gbpath)
{ commitcounter = 0;
  db_to_name = _db_to_name;
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

UTMGBToGB::UTMGBToGB (char const *gbpath,std::string& _db_to_name) : UTMGBBaseQuery (gbpath)
{ commitcounter = 0;
  db_to_name = _db_to_name;
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

UTMGBToGB::UTMGBToGB (char const *gbpath,char const *_db_to_name) : UTMGBBaseQuery (gbpath)
{ commitcounter = 0;
  db_to_name = _db_to_name;
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

UTMGBToGB::~UTMGBToGB ()
{ if (this->db_to->isOpen ()) this->db_to->close ();
  delete this->db_to;
}

void UTMGBToGB::Execute ()
{ this->db_to->open (this->db_to_name.c_str ());
  UTMGBBaseQuery::Execute ();
  this->db_to->commit ();
  this->db_to->close ();
}

int UTMGBToGB::ProcessRow (dbCursor<iptraffic_raw> &tdc)
{ iptraffic_raw rec;

  if (stopped) return 0;
  rec.slink_id = tdc->slink_id;
  rec.srcaddr = tdc->srcaddr;
  rec.dstaddr = tdc->dstaddr;
  rec.nexthop = tdc->nexthop;
  rec.input = tdc->input;
  rec.output = tdc->output;
  rec.dPkts = tdc->dPkts;
  rec.dOctets = tdc->dOctets;
  rec.First = tdc->First;
  rec.Last = tdc->Last;
  rec.srcport = tdc->srcport;
  rec.dstport = tdc->dstport;
  rec.pad = tdc->pad;
  rec.tcp_flags = tdc->tcp_flags;
  rec.prot = tdc->prot;
  rec.tos = tdc->tos;
  rec.src_as = tdc->src_as;
  rec.dst_as = tdc->dst_as;
  rec.src_mask = tdc->src_mask;
  rec.dst_mask = tdc->dst_mask;
  rec.t_class = tdc->t_class;
  rec.account_id = tdc->account_id;
  rec.timestamp = tdc->timestamp;
  this->db_to->insert (rec);
  this->commitcounter++;
  if (this->commitcounter == 10000)
  { this->db_to->commit ();
    this->commitcounter = 0;
  }
  return 1;
}

GBBaseQuery::GBBaseQuery ()
{ rows_count = 0;
  db_from_name = "";
  filter = "";
}

GBBaseQuery::GBBaseQuery (std::string& _db_from_name)
{ rows_count = 0;
  db_from_name = _db_from_name;
  filter = "";
}

GBBaseQuery::GBBaseQuery (char const *_db_from_name)
{ rows_count = 0;
  db_from_name = _db_from_name;
  filter = "";
}

GBBaseQuery::~GBBaseQuery ()
{}

void GBBaseQuery::Execute ()
{ dbDatabase *tdb;
  dbCursor<iptraffic_raw> *tdc;
  dbQuery *tq;

  tdb = new dbDatabase(dbDatabase::dbReadOnly,4096,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
  tdc = new dbCursor<iptraffic_raw> (tdb);
  tq = new dbQuery (this->filter.c_str ());
#ifdef dbException
  try
  {
#endif
    tdb->open (this->db_from_name.c_str ());
    if (tdb->isOpen ())
    { if ((this->rows_count = tdc->select (*tq)) > 0)
      { do
        { if (this->ProcessRow (*tdc) == 0) break;
        }
        while (tdc->next ());
      }
      tdb->close ();
    }
#ifdef dbException
  }
  catch (dbException& e)
  {
  }
#endif
  delete tq;
  delete tdc;
  delete tdb;
}

int GBBaseQuery::ProcessRow (dbCursor<iptraffic_raw> &tdc)
{ return 1;
}

void GBBaseQuery::SetFilter (std::string& filtertext)
{ filter = filtertext;
}

void GBBaseQuery::SetFilter (char const *filtertext)
{ filter = filtertext;
}

GBToGB::GBToGB () : GBBaseQuery ()
{ commitcounter = 0;
  db_to_name = "";
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

GBToGB::GBToGB (std::string& _db_from_name,std::string& _db_to_name) : GBBaseQuery (_db_from_name)
{ commitcounter = 0;
  db_to_name = _db_to_name;
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

GBToGB::GBToGB (std::string& _db_from_name,char const *_db_to_name) : GBBaseQuery (_db_from_name)
{ commitcounter = 0;
  db_to_name = _db_to_name;
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

GBToGB::GBToGB (char const *_db_from_name,std::string& _db_to_name) : GBBaseQuery (_db_from_name)
{ commitcounter = 0;
  db_to_name = _db_to_name;
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

GBToGB::GBToGB (char const *_db_from_name,char const *_db_to_name) : GBBaseQuery (_db_from_name)
{ commitcounter = 0;
  db_to_name = _db_to_name;
  db_to = new dbDatabase(dbDatabase::dbAllAccess,2048,dbDefaultExtensionQuantum,dbDefaultInitIndexSize,0);
}

GBToGB::~GBToGB ()
{ if (this->db_to->isOpen ()) this->db_to->close ();
  delete this->db_to;
}

void GBToGB::Execute ()
{ this->db_to->open (this->db_to_name.c_str ());
  GBBaseQuery::Execute ();
  this->db_to->commit ();
}

int GBToGB::ProcessRow (dbCursor<iptraffic_raw> &tdc)
{ iptraffic_raw rec;

  if (stopped) return 0;
  rec.slink_id = tdc->slink_id;
  rec.srcaddr = tdc->srcaddr;
  rec.dstaddr = tdc->dstaddr;
  rec.nexthop = tdc->nexthop;
  rec.input = tdc->input;
  rec.output = tdc->output;
  rec.dPkts = tdc->dPkts;
  rec.dOctets = tdc->dOctets;
  rec.First = tdc->First;
  rec.Last = tdc->Last;
  rec.srcport = tdc->srcport;
  rec.dstport = tdc->dstport;
  rec.pad = tdc->pad;
  rec.tcp_flags = tdc->tcp_flags;
  rec.prot = tdc->prot;
  rec.tos = tdc->tos;
  rec.src_as = tdc->src_as;
  rec.dst_as = tdc->dst_as;
  rec.src_mask = tdc->src_mask;
  rec.dst_mask = tdc->dst_mask;
  rec.t_class = tdc->t_class;
  rec.account_id = tdc->account_id;
  rec.timestamp = tdc->timestamp;
  this->db_to->insert (rec);
  this->commitcounter++;
  if (this->commitcounter == 10000)
  { this->db_to->commit ();
    this->commitcounter = 0;
  }
  return 1;
}

GBToGBTimeAggregator::GBToGBTimeAggregator () : GBToGB ()
{}

GBToGBTimeAggregator::GBToGBTimeAggregator (std::string& _db_from_name,std::string& _db_to_name) : GBToGB (_db_from_name,_db_to_name)
{}

GBToGBTimeAggregator::GBToGBTimeAggregator (std::string& _db_from_name,char const *_db_to_name) : GBToGB (_db_from_name,_db_to_name)
{}

GBToGBTimeAggregator::GBToGBTimeAggregator (char const *_db_from_name,std::string& _db_to_name) : GBToGB (_db_from_name,_db_to_name)
{}

GBToGBTimeAggregator::GBToGBTimeAggregator (char const *_db_from_name,char const *_db_to_name) : GBToGB (_db_from_name,_db_to_name)
{}

GBToGBTimeAggregator::~GBToGBTimeAggregator ()
{}

void GBToGBTimeAggregator::Execute ()
{ this->InitAggRanges ();
  GBToGB::Execute ();
}

void GBToGBTimeAggregator::InitAggRanges ()
{}

int GBToGBTimeAggregator::ProcessRow (dbCursor<iptraffic_raw> &tdc)
{ iptraffic_raw rec;

  if (stopped) return 0;
  rec.slink_id = tdc->slink_id;
  rec.srcaddr = tdc->srcaddr;
  rec.dstaddr = tdc->dstaddr;
  rec.nexthop = tdc->nexthop;
  rec.input = tdc->input;
  rec.output = tdc->output;
  rec.dPkts = tdc->dPkts;
  rec.dOctets = tdc->dOctets;
  rec.First = tdc->First;
  rec.Last = tdc->Last;
  rec.srcport = tdc->srcport;
  rec.dstport = tdc->dstport;
  rec.pad = tdc->pad;
  rec.tcp_flags = tdc->tcp_flags;
  rec.prot = tdc->prot;
  rec.tos = tdc->tos;
  rec.src_as = tdc->src_as;
  rec.dst_as = tdc->dst_as;
  rec.src_mask = tdc->src_mask;
  rec.dst_mask = tdc->dst_mask;
  rec.t_class = tdc->t_class;
  rec.account_id = tdc->account_id;
  rec.timestamp = tdc->timestamp;
  for (int j = 0;j < this->lowts.size ();j++)
  { if ((tdc->timestamp >= this->lowts[j]) &&
        (tdc->timestamp < this->hights[j]))
    { rec.timestamp = this->lowts[j];
      break;
    }
  }
  this->db_to->insert (rec);
  this->commitcounter++;
  if (this->commitcounter == 10000)
  { this->db_to->commit ();
    this->commitcounter = 0;
  }
  return 1;
}

GBToGBDataAggregator::GBToGBDataAggregator () : GBToGB ()
{}

GBToGBDataAggregator::GBToGBDataAggregator (std::string& _db_from_name,std::string& _db_to_name) : GBToGB (_db_from_name,_db_to_name)
{}

GBToGBDataAggregator::GBToGBDataAggregator (std::string& _db_from_name,char const *_db_to_name) : GBToGB (_db_from_name,_db_to_name)
{}

GBToGBDataAggregator::GBToGBDataAggregator (char const *_db_from_name,std::string& _db_to_name) : GBToGB (_db_from_name,_db_to_name)
{}

GBToGBDataAggregator::GBToGBDataAggregator (char const *_db_from_name,char const *_db_to_name) : GBToGB (_db_from_name,_db_to_name)
{}

GBToGBDataAggregator::~GBToGBDataAggregator ()
{}

void GBToGBDataAggregator::Execute ()
{ this->InitAggBuffer ();
  emptybuf = 1;
  GBToGB::Execute ();
  this->FlushAggBuffer ();
}

int GBToGBDataAggregator::ProcessRow (dbCursor<iptraffic_raw> &tdc)
{ if (stopped) return 0;
  if (emptybuf)
  { this->CopyRowToAggBuffer (tdc);
    emptybuf = 0;
  }
  else
  { if (this->CheckRow (tdc)) this->Summarize (tdc);
    else
    { if (!emptybuf)
      { this->FlushAggBuffer ();
        this->CopyRowToAggBuffer (tdc);
      }
    }
  }
  return 1;
}

void GBToGBDataAggregator::InitAggBuffer ()
{ aggbuf.slink_id = 0;
  aggbuf.srcaddr = 0;
  aggbuf.dstaddr = 0;
  aggbuf.nexthop = 0;
  aggbuf.input = 0;
  aggbuf.output = 0;
  aggbuf.dPkts = 0;
  aggbuf.dOctets = 0;
  aggbuf.First = 0;
  aggbuf.Last = 0;
  aggbuf.srcport = 0;
  aggbuf.dstport = 0;
  aggbuf.pad = 0;
  aggbuf.tcp_flags = 0;
  aggbuf.prot = 0;
  aggbuf.tos = 0;
  aggbuf.src_as = 0;
  aggbuf.dst_as = 0;
  aggbuf.src_mask = 0;
  aggbuf.dst_mask = 0;
  aggbuf.t_class = 0;
  aggbuf.account_id = 0;
  aggbuf.timestamp = 0;
}

void GBToGBDataAggregator::FlushAggBuffer ()
{}

void GBToGBDataAggregator::CopyRowToAggBuffer (dbCursor<iptraffic_raw> &tdc)
{}

void GBToGBDataAggregator::Summarize (dbCursor<iptraffic_raw> &tdc)
{}

int GBToGBDataAggregator::CheckRow (dbCursor<iptraffic_raw> &tdc)
{ return 0;
}
