#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <string>
#include <vector>
#include "globals.h"
#include "params.h"
#include "err_rep.h"
#include "utmdrgw.rh"

proto_entry proto_table[] =
{{1,"ICMP"},{2,"IGMP"},{3,"GGP"},{4,"IPIP"},
 {6,"TCP"},{17,"UDP"},{47,"GRE"},{50,"ESP"},
 {51,"AH"},{-1,NULL}};

FILE *outfile;
IntParameter period_begin;
IntParameter period_end;
IntParameter account_id;
IntParameter slink_id;
IntParameter source_addr;
IntParameter dest_addr;
IntParameter source_port;
IntParameter dest_port;
IntParameter proto;
int my_argc;
char **my_argv;
bool quiet;
bool help_me;
bool noi;
std::string temp_path;
std::string dbpath;
std::string outfile_name;
char class_filter;
char workmode;
std::vector<int> incoming_classes;
std::vector<int> outgoing_classes;
std::vector<int> paid_classes;
bool to_stdout;
char unit;
char report_type;
int agg_step;

void ProtoName (int prot,char *buffer)
{ int j = 0;

  while (proto_table[j].proto_number != -1)
  { if (prot == proto_table[j].proto_number)
    { strcpy (buffer,proto_table[j].proto_text);
      return;
    }
    j++;
  }
  strcpy (buffer,"");
}

static int ParseDateTime (const char *tdt,int values[])
{ int j,digit_buf,digits_count,syntax_error;

  j = 0;
  while (tdt[j] != 0)
  { if (!isspace (tdt[j])) break;
    j++;
  }
  digits_count = 0;
  syntax_error = 0;
  digit_buf = 0;
  do
  { if (isdigit (tdt[j]))
    { if ((j != 0) && (isdigit (tdt[j-1]))) digit_buf *= 10;
      digit_buf += (tdt[j] - '0');
      digits_count++;
      if (digits_count > 2)
      { syntax_error = 1;
        break;
      }
    }
    else
    { if ((tdt[j] == '-') || (tdt[j] == '/') || (tdt[j] == '.') || (tdt[j] == 0))
      { values[0] = digit_buf;
        if (tdt[j] != 0) j++;
        break;
      }
      else
      { syntax_error = 1;
        break;
      }
    }
    j++;
  }
  while (!syntax_error);
  if (!syntax_error)
  { digit_buf = 0;
    digits_count = 0;
    do
    { if (isdigit (tdt[j]))
      { if ((j != 0) && (isdigit (tdt[j-1]))) digit_buf *= 10;
        digit_buf += (tdt[j] - '0');
        digits_count++;
        if (digits_count > 2)
        { syntax_error = 1;
          break;
        }
      }
      else
      { if ((tdt[j] == '-') || (tdt[j] == '/') || (tdt[j] == '.') || (tdt[j] == 0))
        { values[1] = digit_buf;
          if (tdt[j] != 0) j++;
          break;
        }
        else
        { syntax_error = 1;
          break;
        }
      }
      j++;
    }
    while (!syntax_error);
    if (!syntax_error)
    { digit_buf = 0;
      digits_count = 0;
      do
      { if (isdigit (tdt[j]))
        { if ((j != 0) && (isdigit (tdt[j-1]))) digit_buf *= 10;
          digit_buf += (tdt[j] - '0');
          digits_count++;
          if (digits_count > 4)
          { syntax_error = 1;
            break;
          }
        }
        else
        { if (isspace (tdt[j]) || (tdt[j] == 0))
          { if (digits_count == 2) digit_buf += 2000;
            values[2] = digit_buf;
            if (tdt[j] != 0) j++;
            break;
          }
          else
          { syntax_error = 1;
            break;
          }
        }
        j++;
      }
      while (!syntax_error);
      if (!syntax_error)
      { while (tdt[j] != 0)
        { if (!isspace (tdt[j])) break;
          j++;
        }
        digit_buf = 0;
        digits_count = 0;
        do
        { if (isdigit (tdt[j]))
          { if ((j != 0) && (isdigit (tdt[j-1]))) digit_buf *= 10;
            digit_buf += (tdt[j] - '0');
            digits_count++;
            if (digits_count > 2)
            { syntax_error = 1;
              break;
            }
          }
          else
          { if ((tdt[j] == ':') || (tdt[j] == 0))
            { values[3] = digit_buf;
              if (tdt[j] != 0) j++;
              break;
            }
            else
            { syntax_error = 1;
              break;
            }
          }
          j++;
        }
        while (!syntax_error);
        if (!syntax_error)
        { digit_buf = 0;
          digits_count = 0;
          do
          { if (isdigit (tdt[j]))
            { if ((j != 0) && (isdigit (tdt[j-1]))) digit_buf *= 10;
              digit_buf += (tdt[j] - '0');
              digits_count++;
              if (digits_count > 2)
              { syntax_error = 1;
                break;
              }
            }
            else
            { if ((tdt[j] == ':') || (tdt[j] == 0))
              { values[4] = digit_buf;
                if (tdt[j] != 0) j++;
                break;
              }
              else
              { syntax_error = 1;
                break;
              }
            }
            j++;
          }
          while (!syntax_error);
          if (!syntax_error)
          { digit_buf = 0;
            digits_count = 0;
            do
            { if (isdigit (tdt[j]))
              { if ((j != 0) && (isdigit (tdt[j-1]))) digit_buf *= 10;
                digit_buf += (tdt[j] - '0');
                digits_count++;
                if (digits_count > 2)
                { syntax_error = 1;
                  break;
                }
              }
              else
              { if (tdt[j] == 0)
                { values[5] = digit_buf;
                  break;
                }
              }
              j++;
            }
            while (!syntax_error);
          }
        }
      }
    }
  }
  return syntax_error;
}

static int IsLeapYear (int year)
{ return ((((year % 4) == 0) && ((year % 100) != 0)) || ((year % 400) == 0))?1:0;
}

static int DaysInMonth (int month,int year)
{ static int days[12] = {31,28,31,30,31,30,31,31,30,31,30,31};
  int res;

  res = 0;
  if ((month >= 1) && (month <= 12))
  { res = days[month-1];
    if ((month == 2) && IsLeapYear (year)) res++;
  }
  return res;
}

static int IsGoodDateTime (int values[])
{ return ((values[0] >= 1) && (values[0] <= DaysInMonth (values[1],values[2])) &&
          (values[1] >= 1) && (values[1] <= 12) &&
          (values[2] >= 1970) && (values[2] <= 2037) &&
          (values[3] >= 0) && (values[3] <= 23) &&
          (values[4] >= 0) && (values[4] <= 59) &&
          (values[5] >= 0) && (values[5] <= 59));
}

static int ParseInteger (const char *ti,int *value)
{ int j,digit_buf,digits_count,syntax_error;

  j = 0;
  while (ti[j] != 0)
  { if (!isspace (ti[j])) break;
    j++;
  }
  digits_count = 0;
  syntax_error = 0;
  digit_buf = 0;
  do
  { if (isdigit (ti[j]))
    { if ((j != 0) && (isdigit (ti[j-1]))) digit_buf *= 10;
      digit_buf += (ti[j] - '0');
      digits_count++;
      if (digits_count > 10)
      { syntax_error = 1;
        break;
      }
    }
    else
    { if (isspace (ti[j]) || (ti[j] == 0))
      { *value = digit_buf;
        break;
      }
      else
      { syntax_error = 1;
        break;
      }
    }
    j++;
  }
  while (!syntax_error);
  return syntax_error;
}

static int IsGoodInteger (int value)
{ return (value >= 0);
}

static int IsGoodShort (int value)
{ return ((value >= 0) && (value <= 65535));
}

static int ParseIPAddress (const char *tip,int values[])
{ int j,digit_buf,digits_count,syntax_error;

  j = 0;
  while (tip[j] != 0)
  { if (!isspace (tip[j])) break;
    j++;
  }
  digits_count = 0;
  digit_buf = 0;
  syntax_error = 0;
  do
  { if (isdigit (tip[j]))
    { if ((j != 0) && (isdigit (tip[j-1]))) digit_buf *= 10;
      digit_buf += (tip[j] - '0');
      digits_count++;
      if (digits_count > 3)
      { syntax_error = 1;
        break;
      }
    }
    else
    { if (tip[j] == '.')
      { values[0] = digit_buf;
        j++;
        break;
      }
      else
      { syntax_error = 1;
        break;
      }
    }
    j++;
  }
  while (!syntax_error);
  if (!syntax_error)
  { digit_buf = 0;
    digits_count = 0;
    do
    { if (isdigit (tip[j]))
      { if ((j != 0) && (isdigit (tip[j-1]))) digit_buf *= 10;
        digit_buf += (tip[j] - '0');
        digits_count++;
        if (digits_count > 3)
        { syntax_error = 1;
          break;
        }
      }
      else
      { if (tip[j] == '.')
        { values[1] = digit_buf;
          j++;
          break;
        }
        else
        { syntax_error = 1;
          break;
        }
      }
      j++;
    }
    while (!syntax_error);
    if (!syntax_error)
    { digit_buf = 0;
      digits_count = 0;
      do
      { if (isdigit (tip[j]))
        { if ((j != 0) && (isdigit (tip[j-1]))) digit_buf *= 10;
          digit_buf += (tip[j] - '0');
          digits_count++;
          if (digits_count > 3)
          { syntax_error = 1;
            break;
          }
        }
        else
        { if (tip[j] == '.')
          { values[2] = digit_buf;
            j++;
            break;
          }
          else
          { syntax_error = 1;
            break;
          }
        }
        j++;
      }
      while (!syntax_error);
      if (!syntax_error)
      { digit_buf = 0;
        digits_count = 0;
        do
        { if (isdigit (tip[j]))
          { if ((j != 0) && (isdigit (tip[j-1]))) digit_buf *= 10;
            digit_buf += (tip[j] - '0');
            digits_count++;
            if (digits_count > 3)
            { syntax_error = 1;
              break;
            }
          }
          else
          { if (isspace (tip[j]) || (tip[j] == 0))
            { values[3] = digit_buf;
              break;
            }
            else
            { syntax_error = 1;
              break;
            }
          }
          j++;
        }
        while (!syntax_error);
      }
    }
  }
  return syntax_error;
}

static int IsGoodIPAddress (int values[])
{ return ((values[0] >= 0) && (values[0] <= 255) &&
          (values[1] >= 0) && (values[1] <= 255) &&
          (values[2] >= 0) && (values[2] <= 255) &&
          (values[3] >= 0) && (values[3] <= 255));
}

static int ParseProtocol (const char *tp,int *value)
{ int j,digit_buf,sym_count,syntax_error;
  char equal;

  j = 0;
  while (tp[j] != 0)
  { if (!isspace (tp[j])) break;
    j++;
  }
  digit_buf = 0;
  sym_count = 0;
  syntax_error = 0;
  if (isdigit (tp[j]))
  { do
    { if (isdigit (tp[j]))
      { if ((j != 0) && (isdigit (tp[j-1]))) digit_buf *= 10;
        digit_buf += (tp[j] - '0');
        sym_count++;
        if (sym_count > 3)
        { syntax_error = 1;
          break;
        }
      }
      else
      { if (isspace (tp[j]) || (tp[j] == 0))
        { *value = digit_buf;
          break;
        }
      }
      j++;
    }
    while (!syntax_error);
  }
  else
  { if (isalpha (tp[j]))
    { do
      { if (isalpha (tp[j])) sym_count++;
        else
        { syntax_error = 1;
          break;
        }
        j++;
      }
      while (tp[j] != 0);
      if (!syntax_error)
      { for (int k = 0;proto_table[k].proto_number != -1;k++)
        { if (strlen (proto_table[k].proto_text) == sym_count)
          { equal = 1;
            for (int l = 0;l < sym_count;l++)
            { equal &= (toupper (tp[j-sym_count+l]) == proto_table[k].proto_text[l]);
              if (!equal) break;
            }
            if (equal)
            { *value = proto_table[k].proto_number;
              break;
            }
          }
        }
      }
    }
    else syntax_error = 1;
  }
  return syntax_error;
}

static int IsGoodProtocol (int proto)
{ for (int j = 0;proto_table[j].proto_number != -1;j++)
  { if (proto == proto_table[j].proto_number) return 1;
  }
  return 0;
}

static int ParseUnit (const char *tu,char *value)
{ int j,syntax_error;

  j = 0;
  while (tu[j] != 0)
  { if (!isspace (tu[j])) break;
    j++;
  }
  syntax_error = 0;
  if (isalpha (tu[j]))
  { do
    { if (isalpha (tu[j]))
      { switch (toupper (tu[j]))
        { case 'B':
           *value = 1;
           break;
          case 'K':
           *value = 2;
           break;
          case 'M':
           *value = 3;
           break;
          default:
           syntax_error = 1;
           break;
        }
      }
      else
      { if (tu[j] == 0) break;
        if (!isspace (tu[j]))
        { syntax_error = 1;
          break;
        }
      }
      j++;
    }
    while (!syntax_error);
  }
  else
  { if (isdigit (tu[j]))
    { do
      { if (isdigit (tu[j]))
        { switch (tu[j])
          { case '1':
             *value = 1;
             break;
            case '2':
             *value = 2;
             break;
            case '3':
             *value = 3;
             break;
            default:
             syntax_error = 1;
             break;
          }
        }
        else
        { if (tu[j] == 0) break;
          if (!isspace (tu[j]))
          { syntax_error = 1;
            break;
          }
        }
        j++;
      }
      while (!syntax_error);
    }
    else syntax_error = 1;
  }
  return syntax_error;
}

static int ParseReportType (char const *trt,char *value)
{ int j,syntax_error;

  j = 0;
  while (trt[j] != 0)
  { if (!isspace (trt[j])) break;
    j++;
  }
  syntax_error = 0;
  do
  { if (isdigit (trt[j]))
    { switch (trt[j])
      { case '1':
         *value = 1;
         break;
        case '2':
         *value = 2;
         break;
        case '3':
         *value = 3;
         break;
        default:
         syntax_error = 1;
         break;
      }
    }
    else
    { if (trt[j] == 0) break;
      if (!isspace (trt[j]))
      { syntax_error = 1;
        break;
      }
    }
    j++;
  }
  while (!syntax_error);
  return syntax_error;
}

static int ParseClassFilter (char const *cft,char *value)
{ int j,syntax_error;

  j = 0;
  while (cft[j] != 0)
  { if (!isspace (cft[j])) break;
    j++;
  }
  syntax_error = 0;
  do
  { if (isdigit (cft[j]))
    { switch (cft[j])
      { case '1':
         *value = 0;
         break;
        case '2':
         *value = 1;
         break;
        case '3':
         *value = 2;
         break;
        case '4':
         *value = 3;
         break;
        default:
         syntax_error = 1;
         break;
      }
    }
    else
    { if (isalpha (cft[j]))
      { switch (cft[j])
        { case 'n':
           *value = 0;
           break;
          case 'i':
           *value = 1;
           break;
          case 'o':
           *value = 2;
           break;
          case 'p':
           *value = 3;
           break;
          default:
           syntax_error = 1;
           break;
        }
      }
      else
      { if (cft[j] == 0) break;
        if (!isspace (cft[j]))
        { syntax_error = 1;
          break;
        }
      }
    }
    j++;
  }
  while (!syntax_error);
  return syntax_error;
}

static int ParseWorkMode (char const *wmt,char *value)
{ int j,syntax_error;

  j = 0;
  while (wmt[j] != 0)
  { if (!isspace (wmt[j])) break;
    j++;
  }
  syntax_error = 0;
  do
  { if (isdigit (wmt[j]))
    { switch (wmt[j])
      { case '1':
         *value = 0;
         break;
        case '2':
         *value = 1;
         break;
        default:
         syntax_error = 1;
         break;
      }
    }
    else
    { if (isalpha (wmt[j]))
      { switch (wmt[j])
        { case 'r':
           *value = 0;
           break;
          case 'c':
           *value = 1;
           break;
          default:
           syntax_error = 1;
           break;
        }
      }
      else
      { if (wmt[j] == 0) break;
        if (!isspace (wmt[j]))
        { syntax_error = 1;
          break;
        }
      }
    }
    j++;
  }
  while (!syntax_error);
  return syntax_error;
}

static char *GetCmdLineOpt (char const *name)
{ int j,k,l,m;
  char ok;

  if (my_argc > 1)
  { for (j = my_argc - 1;j >= 1;j--)
    { k = strlen (my_argv[j]);
      if ((my_argv[j][0] == '-') && (k > 1))
      { l = 1;
        m = 0;
        while ((my_argv[j][l] != '=') && (my_argv[j][l] != 0))
        { l++;
          m++;
        }
        if (m == strlen (name))
        { l = 1;
          m = 0;
          ok = 1;
          while ((my_argv[j][l] != '=') && (my_argv[j][l] != 0) && (name[m] != 0))
          { ok &= (my_argv[j][l] == name[m]);
            if (!ok) break;
            l++;
            m++;
          }
          if (ok) return (strdup ((l == k - 1)?"":(char *)&(my_argv[j][l+1])));
        }
      }
    }
  }
  return NULL;
}

static bool GetCmdLineSwitch (char const *name)
{ if (my_argc > 1)
  { for (int j = my_argc - 1;j >= 1;j--)
    { if ((my_argv[j][0] == '-') &&
          (strlen (my_argv[j]) > 1) &&
          (strcmp (&(my_argv[j][1]),name) == 0)) return true;
    }
  }
  return false;
}

static void Chomp (char *data)
{ int l;

  if ((data != NULL) && (data[0] != 0))
  { l = strlen (data);
    if (data[l-1] == '\n')
    { if (data[l-2] == '\r') data[l-2] = 0;
      else data[l-1] = 0;
    }
  }
}

static char *Trim (const char *str)
{ int j,startpos,len;
  char *res;

  if (str != NULL)
  { j = strlen (str);
    if (j > 1)
    { j = 0;
      while (str[j] != 0)
      { if (!isspace (str[j])) break;
        j++;
      }
      startpos = j;
      j = strlen (str)-1;
      while (j > startpos)
      { if (!isspace (str[j])) break;
        j--;
      }
      len = j - startpos + 1;
      res = (char *)malloc (len + 1);
      memset (res,0,len + 1);
      strncpy (res,&(str[startpos]),len);
    }
    else res = strdup (str);
    return res;
  }
  return NULL;
}

static void gfw_common_opt (char *buffer,int bufsize,int ckb_id,int txt_id)
{ EnterCriticalSection (&mainwnd_cs);
  if ((mainwnd != 0) && (IsDlgButtonChecked (mainwnd,ckb_id)))
   GetWindowText (GetDlgItem (mainwnd,txt_id),buffer,bufsize);
  else
   strcpy (buffer,"");
  LeaveCriticalSection (&mainwnd_cs);
}

static void gfw_common_txt (char *buffer,int bufsize,int txt_id)
{ EnterCriticalSection (&mainwnd_cs);
  if (mainwnd != 0) GetWindowText (GetDlgItem (mainwnd,txt_id),buffer,bufsize);
  else strcpy (buffer,"");
  LeaveCriticalSection (&mainwnd_cs);
}

static void ptw_common_opt (char *buffer,int bufsize,int ckb_id,int txt_id)
{ HWND swnd;

  EnterCriticalSection (&mainwnd_cs);
  if (mainwnd != 0)
  { CheckDlgButton (mainwnd,ckb_id,BST_CHECKED);
    swnd = GetDlgItem (mainwnd,txt_id);
    EnableWindow (swnd,TRUE);
    SetWindowText (swnd,buffer);
  }
  LeaveCriticalSection (&mainwnd_cs);
}

static void ptw_common_txt (char *buffer,int bufsize,int txt_id)
{ EnterCriticalSection (&mainwnd_cs);
  if (mainwnd != 0) SetWindowText (GetDlgItem (mainwnd,txt_id),buffer);
  LeaveCriticalSection (&mainwnd_cs);
}

static void workmode_gfw (char *buffer,int bufsize)
{ EnterCriticalSection (&mainwnd_cs);
  if (mainwnd != 0)
  { switch (SendDlgItemMessage (mainwnd,IDC_OUT_FORMAT_COMBOBOX,CB_GETCURSEL,0,0))
    { case 0:
       strcpy (buffer,"1");
       break;
      case 1:
       strcpy (buffer,"2");
       break;
      default:
       strcpy (buffer,"");
       break;
    }
  }
  else strcpy (buffer,"");
  LeaveCriticalSection (&mainwnd_cs);
}

static bool workmode_cft (char const *value)
{ char val;

  if (ParseWorkMode (value,&val) == 0)
  { workmode = val;
    return true;
  }
  else return false;
}

static void workmode_setdefault ()
{ ParseWorkMode (DefaultWorkMode,&workmode);
}

static void period_begin_gfw (char *buffer,int bufsize)
{ gfw_common_opt (buffer,bufsize,IDC_PERIOD_BEGIN_CHECKBOX,IDC_PERIOD_BEGIN_TEXT);
}

static void period_begin_ptw (char *buffer,int bufsize)
{ ptw_common_opt (buffer,bufsize,IDC_PERIOD_BEGIN_CHECKBOX,IDC_PERIOD_BEGIN_TEXT);
}

static bool period_begin_datetime_cft (char const *value)
{ int values[6];
  struct tm tmbuf;

  if ((ParseDateTime (value,values) == 0) && (IsGoodDateTime (values)))
  { memset (&tmbuf,0,sizeof (tmbuf));
    tmbuf.tm_sec = values[5];
    tmbuf.tm_min = values[4];
    tmbuf.tm_hour = values[3];
    tmbuf.tm_mday = values[0];
    tmbuf.tm_mon = values[1]-1;
    tmbuf.tm_year = values[2]-1900;
    tmbuf.tm_isdst = -1;
    period_begin = (int)mktime (&tmbuf);
    return true;
  }
  else return false;
}

static void period_begin_datetime_ctt (char *buffer,int bufsize)
{ time_t ts;

  if (!period_begin.IsEmpty ())
  { ts = (time_t)(period_begin.Value ());
    strftime (buffer,bufsize,"%d.%m.%Y %H:%M:%S",localtime (&ts));
  }
  else strcpy (buffer,"");
}

static bool period_begin_timestamp_cft (char const *value)
{ int val;

  if ((ParseInteger (value,&val) == 0) && (IsGoodInteger (val)))
  { period_begin = val;
    return true;
  }
  else return false;
}

static void period_end_gfw (char *buffer,int bufsize)
{ gfw_common_opt (buffer,bufsize,IDC_PERIOD_END_CHECKBOX,IDC_PERIOD_END_TEXT);
}

static void period_end_ptw (char *buffer,int bufsize)
{ ptw_common_opt (buffer,bufsize,IDC_PERIOD_END_CHECKBOX,IDC_PERIOD_END_TEXT);
}

static bool period_end_datetime_cft (char const *value)
{ int values[6];
  struct tm tmbuf;

  if ((ParseDateTime (value,values) == 0) && (IsGoodDateTime (values)))
  { memset (&tmbuf,0,sizeof (tmbuf));
    tmbuf.tm_sec = values[5];
    tmbuf.tm_min = values[4];
    tmbuf.tm_hour = values[3];
    tmbuf.tm_mday = values[0];
    tmbuf.tm_mon = values[1]-1;
    tmbuf.tm_year = values[2]-1900;
    tmbuf.tm_isdst = -1;
    period_end = (int)mktime (&tmbuf);
    return true;
  }
  else return false;
}

static bool period_end_timestamp_cft (char const *value)
{ int val;

  if ((ParseInteger (value,&val) == 0) && (IsGoodInteger (val)))
  { period_end = val;
    return true;
  }
  else return false;
}

static void period_end_datetime_ctt (char *buffer,int bufsize)
{ time_t ts;

  if (!period_end.IsEmpty ())
  { ts = (time_t)(period_end.Value ());
    strftime (buffer,bufsize,"%d.%m.%Y %H:%M:%S",localtime (&ts));
  }
  else strcpy (buffer,"");
}

static void path_gfw (char *buffer,int bufsize)
{ gfw_common_txt (buffer,bufsize,IDC_PATH_TO_DATA_TEXT);
}

static void path_ptw (char *buffer,int bufsize)
{ ptw_common_txt (buffer,bufsize,IDC_PATH_TO_DATA_TEXT);
}

static bool path_cft (char const *value)
{ dbpath = value;
  if ((dbpath.size () > 0) && (dbpath[dbpath.size () - 1] != PathDivider)) dbpath += PathDivider;
  return true;
}

static void path_ctt (char *buffer,int bufsize)
{ strcpy (buffer,dbpath.c_str ());
}

static void path_setdefault ()
{ dbpath = DefaultDataFilePath;
}

static void account_id_gfw (char *buffer,int bufsize)
{ gfw_common_opt (buffer,bufsize,IDC_ACCOUNT_ID_CHECKBOX,IDC_ACCOUNT_ID_TEXT);
}

static void account_id_ptw (char *buffer,int bufsize)
{ ptw_common_opt (buffer,bufsize,IDC_ACCOUNT_ID_CHECKBOX,IDC_ACCOUNT_ID_TEXT);
}

static bool account_id_cft (char const *value)
{ int val;

  if ((ParseInteger (value,&val) == 0) && (IsGoodInteger (val)))
  { account_id = val;
    return true;
  }
  else return false;
}

static void account_id_ctt (char *buffer,int bufsize)
{ if (!account_id.IsEmpty ()) sprintf (buffer,"%d",account_id.Value ());
  else strcpy (buffer,"");
}

static void slink_id_gfw (char *buffer,int bufsize)
{ gfw_common_opt (buffer,bufsize,IDC_SLINK_ID_CHECKBOX,IDC_SLINK_ID_TEXT);
}

static void slink_id_ptw (char *buffer,int bufsize)
{ ptw_common_opt (buffer,bufsize,IDC_SLINK_ID_CHECKBOX,IDC_SLINK_ID_TEXT);
}

static bool slink_id_cft (char const *value)
{ int val;

  if ((ParseInteger (value,&val) == 0) && (IsGoodInteger (val)))
  { slink_id = val;
    return true;
  }
  else return false;
}

static void slink_id_ctt (char *buffer,int bufsize)
{ if (!slink_id.IsEmpty ()) sprintf (buffer,"%d",slink_id.Value ());
  else strcpy (buffer,"");
}

static void class_filter_gfw (char *buffer,int bufsize)
{ EnterCriticalSection (&mainwnd_cs);
  if ((mainwnd != 0) && (IsDlgButtonChecked (mainwnd,IDC_TRAF_TYPE_CHECKBOX)))
  { switch (SendDlgItemMessage (mainwnd,IDC_TRAF_TYPE_COMBOBOX,CB_GETCURSEL,0,0))
    { case 0:
       strcpy (buffer,"2");
       break;
      case 1:
       strcpy (buffer,"3");
       break;
      case 2:
       strcpy (buffer,"4");
       break;
      default:
       strcpy (buffer,"");
       break;
    }
  }
  else strcpy (buffer,"");
  LeaveCriticalSection (&mainwnd_cs);
}

static void class_filter_ptw (char *buffer,int bufsize)
{ EnterCriticalSection (&mainwnd_cs);
  switch (buffer[0])
  { case '2':
    case '3':
    case '4':
     CheckDlgButton (mainwnd,IDC_TRAF_TYPE_CHECKBOX,BST_CHECKED);
     EnableWindow (GetDlgItem (mainwnd,IDC_TRAF_TYPE_COMBOBOX),TRUE);
     SendDlgItemMessage (mainwnd,IDC_TRAF_TYPE_COMBOBOX,CB_SETCURSEL,buffer[0] - '1',0);
     break;
    default:
     CheckDlgButton (mainwnd,IDC_TRAF_TYPE_CHECKBOX,BST_UNCHECKED);
     EnableWindow (GetDlgItem (mainwnd,IDC_TRAF_TYPE_COMBOBOX),FALSE);
     break;
  }
  LeaveCriticalSection (&mainwnd_cs);
}

static bool class_filter_cft (char const *value)
{ char val;

  if (ParseClassFilter (value,&val) == 0)
  { class_filter = val;
    return true;
  }
  else return false;
}

static void class_filter_ctt (char *buffer,int bufsize)
{ switch (class_filter)
  { case 1:
    case 2:
    case 3:
    case 4:
     buffer[0] = class_filter + '0';
     buffer[1] = 0;
     break;
    default:
     buffer[0] = 0;
     break;
  }
}

static void class_filter_setdefault ()
{ class_filter = 0;
}

static void source_addr_gfw (char *buffer,int bufsize)
{ gfw_common_opt (buffer,bufsize,IDC_SOURCE_ADDR_CHECKBOX,IDC_SOURCE_ADDR_TEXT);
}

static void source_addr_ptw (char *buffer,int bufsize)
{ ptw_common_opt (buffer,bufsize,IDC_SOURCE_ADDR_CHECKBOX,IDC_SOURCE_ADDR_TEXT);
}

static bool source_addr_cft (char const *value)
{ int values[4];

  if ((ParseIPAddress (value,values) == 0) && (IsGoodIPAddress (values)))
  { source_addr = (values[0] << 24) | (values[1] << 16) | (values[2] << 8) | values[3];
    return true;
  }
  else return false;
}

static void source_addr_ctt (char *buffer,int bufsize)
{ int values[4];
  int v;

  if (!source_addr.IsEmpty ())
  { v = source_addr.Value ();
    values[0] = (v >> 24) & 0xFF;
    values[1] = (v >> 16) & 0xFF;
    values[2] = (v >> 8) & 0xFF;
    values[3] = v & 0xFF;
    sprintf (buffer,"%d.%d.%d.%d",values[0],values[1],values[2],values[3]);
  }
  else strcpy (buffer,"");
}

static void dest_addr_gfw (char *buffer,int bufsize)
{ gfw_common_opt (buffer,bufsize,IDC_DEST_ADDR_CHECKBOX,IDC_DEST_ADDR_TEXT);
}

static void dest_addr_ptw (char *buffer,int bufsize)
{ ptw_common_opt (buffer,bufsize,IDC_DEST_ADDR_CHECKBOX,IDC_DEST_ADDR_TEXT);
}

static bool dest_addr_cft (char const *value)
{ int values[4];

  if ((ParseIPAddress (value,values) == 0) && (IsGoodIPAddress (values)))
  { dest_addr = (values[0] << 24) | (values[1] << 16) | (values[2] << 8) | values[3];
    return true;
  }
  else return false;
}

static void dest_addr_ctt (char *buffer,int bufsize)
{ int values[4];
  int v;

  if (!dest_addr.IsEmpty ())
  { v = dest_addr.Value ();
    values[0] = (v >> 24) & 0xFF;
    values[1] = (v >> 16) & 0xFF;
    values[2] = (v >> 8) & 0xFF;
    values[3] = v & 0xFF;
    sprintf (buffer,"%d.%d.%d.%d",values[0],values[1],values[2],values[3]);
  }
  else strcpy (buffer,"");
}

static void source_port_gfw (char *buffer,int bufsize)
{ gfw_common_opt (buffer,bufsize,IDC_SOURCE_PORT_CHECKBOX,IDC_SOURCE_PORT_TEXT);
}

static void source_port_ptw (char *buffer,int bufsize)
{ ptw_common_opt (buffer,bufsize,IDC_SOURCE_PORT_CHECKBOX,IDC_SOURCE_PORT_TEXT);
}

static bool source_port_cft (char const *value)
{ int val;

  if ((ParseInteger (value,&val) == 0) && (IsGoodShort (val)))
  { source_port = val;
    return true;
  }
  else return false;
}

static void source_port_ctt (char *buffer,int bufsize)
{ if (!source_port.IsEmpty ()) sprintf (buffer,"%d",source_port.Value ());
  else strcpy (buffer,"");
}

static void dest_port_gfw (char *buffer,int bufsize)
{ gfw_common_opt (buffer,bufsize,IDC_DEST_PORT_CHECKBOX,IDC_DEST_PORT_TEXT);
}

static void dest_port_ptw (char *buffer,int bufsize)
{ ptw_common_opt (buffer,bufsize,IDC_DEST_PORT_CHECKBOX,IDC_DEST_PORT_TEXT);
}

static bool dest_port_cft (char const *value)
{ int val;

  if ((ParseInteger (value,&val) == 0) && (IsGoodShort (val)))
  { dest_port = val;
    return true;
  }
  else return false;
}

static void dest_port_ctt (char *buffer,int bufsize)
{ if (!dest_port.IsEmpty ()) sprintf (buffer,"%d",dest_port.Value ());
  else strcpy (buffer,"");
}

static void protocol_gfw (char *buffer,int bufsize)
{ gfw_common_opt (buffer,bufsize,IDC_PROTO_CHECKBOX,IDC_PROTO_TEXT);
}

static void protocol_ptw (char *buffer,int bufsize)
{ ptw_common_opt (buffer,bufsize,IDC_PROTO_CHECKBOX,IDC_PROTO_TEXT);
}

static bool protocol_cft (char const *value)
{ int val;

  if ((ParseProtocol (value,&val) == 0) && (IsGoodProtocol (val)))
  { proto = val;
    return true;
  }
  else return false;
}

static void protocol_ctt (char *buffer,int bufsize)
{ int v;

  if (!proto.IsEmpty ())
  { v = proto.Value ();
    ProtoName (v,buffer);
    if (buffer[0] == 0) sprintf (buffer,"%d",v);
  }
  else strcpy (buffer,"");
}

static int protocol_skip ()
{ return (workmode == 0)?0:3;
}

static void report_type_gfw (char *buffer,int bufsize)
{ EnterCriticalSection (&mainwnd_cs);
  if (mainwnd != 0)
  { switch (SendDlgItemMessage (mainwnd,IDC_REPORT_TYPE_COMBOBOX,CB_GETCURSEL,0,0))
    { case 0:
       strcpy (buffer,"1");
       break;
      case 1:
       strcpy (buffer,"2");
       break;
      case 2:
       strcpy (buffer,"3");
       break;
      default:
       strcpy (buffer,"");
       break;
    }
  }
  else strcpy (buffer,"");
  LeaveCriticalSection (&mainwnd_cs);
}

static void report_type_ptw (char *buffer,int bufsize)
{ EnterCriticalSection (&mainwnd_cs);
  if (mainwnd != 0)
  { switch (buffer[0])
    { case '1':
      case '2':
      case '3':
       SendDlgItemMessage (mainwnd,IDC_REPORT_TYPE_COMBOBOX,CB_SETCURSEL,buffer[0] - '1',0);
       if (buffer[0] == '2') EnableWindow (GetDlgItem (mainwnd,IDC_AGGSTEP_TEXT),TRUE);
       break;
      default:
       break;
    }
  }
  LeaveCriticalSection (&mainwnd_cs);
}

static bool report_type_cft (char const *value)
{ if (ParseReportType (value,&report_type) != 0) report_type = DefaultReportType;
  return true;
}

static void report_type_ctt (char *buffer,int bufsize)
{ switch (report_type)
  { case 1:
    case 2:
    case 3:
     buffer[0] = report_type + '0';
     buffer[1] = 0;
     break;
    default:
     buffer[0] = 0;
     break;
  }
}

static void report_type_setdefault ()
{ report_type = DefaultReportType;
}

static int report_type_skip ()
{ return (report_type == 2)?0:1;
}

static void aggregation_step_gfw (char *buffer,int bufsize)
{ gfw_common_txt (buffer,bufsize,IDC_AGGSTEP_TEXT);
}

static void aggregation_step_ptw (char *buffer,int bufsize)
{ ptw_common_txt (buffer,bufsize,IDC_AGGSTEP_TEXT);
}

static bool aggregation_step_cft (char const *value)
{ int val;

  if (report_type == 2)
  { if ((ParseInteger (value,&val) == 0) && (IsGoodShort (val)))
    { agg_step = val;
      return true;
    }
    else return false;
  }
  else return true;
}

static void aggregation_step_ctt (char *buffer,int bufsize)
{ sprintf (buffer,"%d",agg_step);
}

static void aggregation_step_setdefault ()
{ if (report_type == 2) ParseInteger (DefaultAggregationStep,&agg_step);
}

static void unit_gfw (char *buffer,int bufsize)
{ EnterCriticalSection (&mainwnd_cs);
  if (mainwnd != 0)
  { switch (SendDlgItemMessage (mainwnd,IDC_TRAF_UNIT_COMBOBOX,CB_GETCURSEL,0,0))
    { case 0:
       strcpy (buffer,"b");
       break;
      case 1:
       strcpy (buffer,"k");
       break;
      case 2:
       strcpy (buffer,"m");
       break;
      default:
       strcpy (buffer,"");
       break;
    }
  }
  else strcpy (buffer,"");
  LeaveCriticalSection (&mainwnd_cs);
}

static void unit_ptw (char *buffer,int bufsize)
{ int num;

  EnterCriticalSection (&mainwnd_cs);
  if (mainwnd != 0)
  { switch (toupper (buffer[0]))
    { case 'B':
      case '1':
       num = 0;
       break;
      case 'K':
      case '2':
       num = 1;
       break;
      case 'M':
      case '3':
      default:
       num = 2;
       break;
    }
    SendDlgItemMessage (mainwnd,IDC_TRAF_UNIT_COMBOBOX,CB_SETCURSEL,num,0);
  }
  LeaveCriticalSection (&mainwnd_cs);
}

static bool unit_cft (char const *value)
{ return (ParseUnit (value,&unit) == 0);
}

static void unit_ctt (char *buffer,int bufsize)
{ switch (unit)
  { case 1:
    case 2:
    case 3:
     buffer[0] = unit + '0';
     buffer[1] = 0;
     break;
    default:
     buffer[0] = 0;
     break;
  }
}

static void unit_setdefault ()
{ ParseUnit (DefaultUnit,&unit);
}

static void outfile_gfw (char *buffer,int bufsize)
{ gfw_common_txt (buffer,bufsize,IDC_OUTFILE_TEXT);
}

static void outfile_ptw (char *buffer,int bufsize)
{ ptw_common_txt (buffer,bufsize,IDC_OUTFILE_TEXT);
}

static bool outfile_cft (char const *value)
{ outfile_name = value;
  return true;
}

static void outfile_ctt (char *buffer,int bufsize)
{ int m;

  m = (outfile_name.size () > (bufsize-1))?outfile_name.size ():bufsize-1;
  strncpy (buffer,outfile_name.c_str (),m);
  buffer[m+1] = 0;
}

static void outfile_setdefault ()
{ switch (workmode)
  { case 0:
     outfile_name = DefaultRepModeResultFile;
     break;
    case 1:
     outfile_name = DefaultCvtModeResultFile;
     break;
    default:
     break;
  }
}

typedef struct
{ char *opt_long;
  char *opt_short;
  int namestr_id;
  void (*get_from_window)(char *,int);
  void (*put_to_window)(char *,int);
  bool (*convert_from_text)(char const *);
  void (*convert_to_text)(char *,int);
  void (*setdefault)();
  int (*skip)();
} param;

static param params[] =
{{"mode",NULL,0,workmode_gfw,NULL,workmode_cft,NULL,workmode_setdefault,NULL},
 {"bp",NULL,IDS_NAME_PERIOD_BEGIN,period_begin_gfw,period_begin_ptw,period_begin_datetime_cft,period_begin_datetime_ctt,NULL,NULL},
 {"ep",NULL,IDS_NAME_PERIOD_END,period_end_gfw,period_end_ptw,period_end_datetime_cft,period_end_datetime_ctt,NULL,NULL},
 {"bt",NULL,0,NULL,period_begin_ptw,period_begin_timestamp_cft,period_begin_datetime_ctt,NULL,NULL},
 {"et",NULL,0,NULL,period_end_ptw,period_end_timestamp_cft,period_end_datetime_ctt,NULL,NULL},
 {"path",NULL,0,path_gfw,path_ptw,path_cft,path_ctt,path_setdefault,NULL},
 {"a",NULL,IDS_NAME_ACCOUNT_ID,account_id_gfw,account_id_ptw,account_id_cft,account_id_ctt,NULL,NULL},
 {"s",NULL,0,slink_id_gfw,slink_id_ptw,slink_id_cft,slink_id_ctt,NULL,NULL},
 {"cf",NULL,0,class_filter_gfw,class_filter_ptw,class_filter_cft,class_filter_ctt,class_filter_setdefault,NULL},
 {"srcaddr","sa",IDS_NAME_SOURCE_ADDR,source_addr_gfw,source_addr_ptw,source_addr_cft,source_addr_ctt,NULL,NULL},
 {"dstaddr","da",IDS_NAME_DEST_ADDR,dest_addr_gfw,dest_addr_ptw,dest_addr_cft,dest_addr_ctt,NULL,NULL},
 {"srcport","sp",IDS_NAME_SOURCE_PORT,source_port_gfw,source_port_ptw,source_port_cft,source_port_ctt,NULL,NULL},
 {"dstport","dp",IDS_NAME_DEST_PORT,dest_port_gfw,dest_port_ptw,dest_port_cft,dest_port_ctt,NULL,NULL},
 {"proto","pr",IDS_NAME_PROTO,protocol_gfw,protocol_ptw,protocol_cft,protocol_ctt,NULL,protocol_skip},
 {"report","rep",0,report_type_gfw,report_type_ptw,report_type_cft,report_type_ctt,report_type_setdefault,report_type_skip},
 {"aggstep","astp",IDS_NAME_AGGSTEP,aggregation_step_gfw,aggregation_step_ptw,aggregation_step_cft,aggregation_step_ctt,aggregation_step_setdefault,NULL},
 {"unit","u",0,unit_gfw,unit_ptw,unit_cft,unit_ctt,unit_setdefault,NULL},
 {"outfile","out",0,outfile_gfw,outfile_ptw,outfile_cft,outfile_ctt,outfile_setdefault,NULL},
 {NULL,NULL,0,NULL,NULL,NULL,NULL,NULL,NULL}};

bool LoadParametersFromWindow ()
{ int j;
  char *par, *tpar, *line;
  bool ok,res;

  res = true;
  line = (char *)malloc (1024);
  for (j = 0;params[j].opt_long != NULL;j++)
  { par = GetCmdLineOpt (params[j].opt_long);
    if ((par == NULL) && (params[j].opt_short != NULL)) par = GetCmdLineOpt (params[j].opt_short);
    if (par != NULL)
    { tpar = Trim (par);
      free (par);
      if ((tpar[0] != 0) && (params[j].convert_from_text != NULL)) params[j].convert_from_text (tpar);
      if ((tpar[0] == 0) && (params[j].setdefault != NULL)) params[j].setdefault ();
      free (tpar);
    }
    else
    { if (params[j].get_from_window != NULL)
      { if (stopped) break;
        line[0] = 0;
        ok = true;
        params[j].get_from_window (line,1023);
        tpar = Trim (line);
        if ((tpar[0] != 0) && (params[j].convert_from_text != NULL))
        { ok = params[j].convert_from_text (tpar);
          free (tpar);
        }
        else
        { free (tpar);
          if (params[j].setdefault != NULL) params[j].setdefault ();
        }
        if (!ok)
        { if (noi)
          { if (params[j].setdefault != NULL) params[j].setdefault ();
          }
          else
          { if (params[j].namestr_id != 0)
            { ParamErrorBox (params[j].namestr_id);
              res = false;
              break;
            }
            else
            { if (params[j].setdefault != NULL) params[j].setdefault ();
            }
          }
        }
      }
      else
      { if (params[j].setdefault != NULL) params[j].setdefault ();
      }
      if (params[j].skip != NULL) j += params[j].skip ();
    }
  }
  free (line);
  return res;
}

void LoadParametersToWindow ()
{ int j;
  char *par, *tpar, *line;
  bool ok;

  line = (char *)malloc (1024);
  for (j = 0;params[j].opt_long != NULL;j++)
  { if (stopped) break;
    par = GetCmdLineOpt (params[j].opt_long);
    if ((par == NULL) && (params[j].opt_short != NULL)) par = GetCmdLineOpt (params[j].opt_short);
    if (par != NULL)
    { tpar = Trim (par);
      free (par);
      if ((tpar[0] != 0) && (params[j].convert_from_text != NULL)) ok = params[j].convert_from_text (tpar);
      if (((tpar[0] == 0) || (!ok)) && (params[j].setdefault != NULL)) params[j].setdefault ();
      if (params[j].convert_to_text != NULL) params[j].convert_to_text (line,1024);
      if (params[j].put_to_window != NULL) params[j].put_to_window (line,1024);
      free (tpar);
      if (params[j].skip != NULL) j += params[j].skip ();
    }
  }
  free (line);
}

void LoadCmdLineFlags ()
{ quiet = GetCmdLineSwitch ("q");
  noi = GetCmdLineSwitch ("noi");
}

static int ParseIntArray (char const *text,std::vector<int>& values)
{ int j,digit_buf,digits_count,syntax_error,range_start;
  char empty_buf,range_mode;

  j = 0;
  while (isspace (text[j])) j++;
  while (text[j] != '=') j++;
  j++;
  if (text[j] == 0) return 0;
  while (isspace (text[j])) j++;
  syntax_error = 0;
  empty_buf = 1;
  range_mode = 0;
  range_start = 0;
  digit_buf = 0;
  digits_count = 0;
  do
  { if (isdigit (text[j]))
    { empty_buf = 0;
      if ((j != 0) && (isdigit (text[j-1]))) digit_buf *= 10;
      digit_buf += (text[j] - '0');
      digits_count++;
      if (digits_count > 10)
      { syntax_error = 1;
        break;
      }
    }
    else
    { if ((text[j] == ',') || (text[j] == 0) && (!empty_buf))
      { if (!range_mode) values.insert (values.end (),digit_buf);
        else
        { if (range_start < digit_buf)
           for (int k = range_start;k <= digit_buf;k++) values.insert (values.end (),k);
          else
           for (int k = digit_buf;k >= range_start;k--) values.insert (values.end (),k);
        }
        empty_buf = 1;
        digit_buf = 0;
        digits_count = 0;
        if (text[j] == 0) break;
      }
      else
      { if (text[j] == '-')
        { range_mode = 1;
          range_start = digit_buf;
          empty_buf = 1;
          digit_buf = 0;
          digits_count = 0;
        }
        else
        { if (text[j] == 0) break;
          if (!isspace (text[j]))
          { syntax_error = 1;
            break;
          }
        }
      }
    }
    j++;
  }
  while (!syntax_error);
  return syntax_error;
}

static int ParseString (char const *text,std::string& str)
{ int j;
  bool quoted;

  j = 0;
  while (isspace (text[j])) j++;
  while (text[j] != '=') j++;
  j++;
  if (text[j] == 0) return 0;
  str.clear ();
  quoted = false;
  while (isspace (text[j])) j++;
  if (text[j] == '"')
  { quoted = true;
    j++;
  }
  do
  { if ((quoted) && (text[j] == '"')) break;
    if ((!quoted) && (isspace (text[j]))) break;
    str.insert (str.end (),text[j]);
    j++;
  }
  while (text[j] != 0);
  return 0;
}

static int incoming_classes_parse (char const *text)
{ incoming_classes.erase (incoming_classes.begin (),incoming_classes.end ());
  return ParseIntArray (text,incoming_classes);
}

static int outgoing_classes_parse (char const *text)
{ outgoing_classes.erase (outgoing_classes.begin (),outgoing_classes.end ());
  return ParseIntArray (text,outgoing_classes);
}

static int paid_classes_parse (char const *text)
{ paid_classes.erase (paid_classes.begin (),paid_classes.end ());
  return ParseIntArray (text,paid_classes);
}

static int temp_path_parse (char const *text)
{ temp_path.clear ();
  return ParseString (text,temp_path);
}

typedef struct
{ char *name;
  int (*parser)(char const *);
} cfgfile_param;

static cfgfile_param config_params[] =
{{"incoming_classes",incoming_classes_parse},
 {"outgoing_classes",outgoing_classes_parse},
 {"paid_classes",paid_classes_parse},
 {"temp_path",temp_path_parse},
 {NULL,NULL}};

void LoadParametersFromConfigFile ()
{ FILE *cf;
  char buffer[512];
  char *line;
  int j;

  j = strlen (my_argv[0]) - 1;
  while ((my_argv[0][j] != PathDivider) && (j > 0)) j--;
  memset (&buffer,0,512);
  if (j > 0) strncpy (&(buffer[0]),my_argv[0],j+1);
  else
  { buffer[0] = '.';
    buffer[1] = PathDivider;
    buffer[2] = 0;
  }
  strcat (&(buffer[0]),"utmdrg.cnf");
  cf = fopen (&(buffer[0]),"r");
  if (cf != NULL)
  { while (!feof (cf))
    { buffer[0] = 0;
      fgets (&(buffer[0]),511,cf);
      Chomp (&(buffer[0]));
      line = Trim (&(buffer[0]));
      if (line[0] == 0) free (line);
      else
      { if (line[0] != '#')
        { j = 0;
          while (config_params[j].name != NULL)
          { if (strncmp (line,config_params[j].name,strlen (config_params[j].name)) == 0)
             config_params[j].parser (line);
            j++;
          }
        }
        free (line);
      }
    }
    fclose (cf);
  }
}
